unit TomasAnimation;

interface

uses
  Windows,
  Messages,
  SysUtils,
  Classes,
  Graphics,
  Controls,
  Forms,
  Dialogs,
  gmp,
  Math,
  TomCam,
  ExtCtrls,
  StdCtrls,
  ComCtrls,
  Buttons;

type

  PCompLink = ^TA_CompLinkRecord;
  PObjLink = ^TA_ObjLinkrecord;

  TBackGround = Procedure;

  TTomasAnimationForm = class(TForm)
    Panel1: TPanel;
    GMPArea: TGMP;
    TrackBar1: TTrackBar;
    MinMaxLabel: TLabel;
    SpeedLabel: TLabel;
    Bevel1: TBevel;
    Bevel2: TBevel;
    CameraSpeedButton: TSpeedButton;
    procedure GoCreateAnimationForm(Sender: TObject);
    procedure Refresh(Sender: TObject);
    procedure GoShowData(Canvas: TCanvas);
    procedure CheckPosition(Sender: TObject);
    procedure CheckCameraMovement(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure CheckCameraMovementEnd(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure OpenCameraForm(Sender: TObject);
  private
    { Private declarations }
    TA_Comps: PCompLink;
    TA_Objects: PObjLink;
    TA_SortList: TList;
    TA_RenderAll: Boolean;
    TA_BGColor: TColor;
    Procedure SortOnZOrder;
  public
    { Public declarations }
    PBackGround: TbackGround;
    KeyIsDown: Boolean;
    Procedure SetCameraPosition(X,Y,Z: Single);
    Procedure GetCameraPosition(Var X,Y,Z: Single);
    Procedure SetCameraRotation(X,Y,Z: Single);
    Procedure GetCameraRotation(Var X,Y,Z: Single);
    Procedure SetLightSourcePosition(X,Y,Z: Single);
    Procedure SetBackGround(BackGround: TPicture);
    Procedure SetBackGroundColor(BGColor: TColor);
    Procedure DrawLine(X1,Y1,Z1,X2,Y2,Z2: Single; Color: TColor);
    procedure SetSpeedPosition(X: Double);
    Procedure GoResumeSimulation;
  end;


PCA_Object = ^TCAppObject;

TCAppObject =
  Record
  TCA_ID: Byte;
  TCA_Object: TObject3D;
  TCA_Center: TPoint3D;
  TCA_RenderMode: TRenderMode;
  TCA_XOff: Single;
  TCA_YOff: Single;
  TCA_ZOff: Single;
  TCA_Color: TColor;
  TCA_Next: PCA_Object;
  TCA_Pred: PCA_Object;
  End;

TCAppearance = class
  Private
  TA_ObjectLink: PCA_Object;
  TA_Position: TPoint3D;
  TA_AngleX: Single;
  TA_AngleY: Single;
  TA_AngleZ: Single;
  TA_Visible: Boolean;
  TA_Link: PCompLink;
  TA_Parent: TCAppearance;
  TA_Dependents: PCompLink;
  TA_ToRender: Boolean;
  Published
  Constructor Create;
  Destructor Destroy; Override;
  Procedure LoadFromFile(FileName: String);
  Procedure Show(X,Y,Z,Angle: Single);
  Procedure Hide;
  Procedure GetPosition(Var X,Y,Z: Single);
  Function GetAngleX: Single;
  Function GetAngleY: Single;
  Function GetAngleZ: Single;
  Procedure SetRenderMode(rm: TRenderMode);
  Procedure SetColor(Color: TColor);
  Procedure AddBoxObject(X,Y,Z: Single; Rm: TRenderMode;
   Color: TColor; Rx,Ry,Rz: Integer; Angle: Single; Dx,Dy,Dz: Single);
  Procedure AddCircleObject(R: Single; Rm: TRenderMode;
   Color: TColor; Rx,Ry,Rz: Integer; Angle: Single; Dx,Dy,Dz: Single);
  Procedure AddCylinderObject(H,R: Single; Rm: TRenderMode;
   Color: TColor; Rx,Ry,Rz: Integer; Angle: Single; Dx,Dy,Dz: Single);
  Procedure Add(TCApp: TCAppearance; X,Y,Z: Single);
  Procedure UnTie;
  End;

TCObject = class
  Private
  TO_Object: TObject3D;
  TO_RenderMode: TRenderMode;
  TO_Color: TColor;
  TO_Position: TPoint3D;
  TO_Angle: Single;
  TO_Link: PObjLink;
  TO_ToRender: Boolean;
  Published
  Constructor Create;
  Destructor Destroy; Override;
  Procedure ReadAppearanceFromFile(FileName: String);
  Procedure Show(X,Y,Z,Angle: Single);
  Procedure SetRenderMode(Rm: TRenderMode);
  End;

TA_CompLinkRecord =
  Record
  TL_Comp: TCAppearance;
  TL_Next: PCompLink;
  TL_Pred: PCompLink;
  End;

TA_ObjLinkRecord =
  Record
  TO_ID: Byte;
  TO_Object: TCObject;
  TO_Next: PObjLink;
  TO_Pred: PObjLink;
  End;


var
  TomasAnimationForm: TTomasAnimationForm;
  TScreenString: String;

implementation

{$R *.DFM}

Uses
  Tomas;

var
  CurrentKeyDown: Word;
  CurrentShiftState: TShiftState;



procedure TTomasAnimationForm.GoCreateAnimationForm(Sender: TObject);
Var
  Light: TPoint3D;
  Direction: TPoint3D;
begin
New(TA_Comps);
TA_Comps^.TL_Comp:=Nil;
TA_Comps^.TL_Next:=TA_Comps;
TA_Comps^.TL_Pred:=TA_Comps;
New(TA_Objects);
TA_Objects^.TO_Next:=TA_Objects;
TA_Objects^.TO_Pred:=TA_Objects;
TA_SortList:=TList.Create;
Light.Z := 100;
Light.Y := 2;
Light.X := 100;
Direction.X := 0;
Direction.Y := 0;
Direction.Z := 0;
GMPArea.SetLightSourcePosition(Light, Direction);
GMPArea.CameraPosition.X:=0;
GMPArea.CameraPosition.Y:=1;
GMPArea.CameraPosition.Z:=10;
TScreenString:='';
TA_RenderAll:=TRUE;
KeyIsDown:=FALSE;
end;

Procedure TTomasAnimationForm.SetCameraPosition(X,Y,Z: Single);
  Begin           
  GMPArea.CameraPosition.X:=X;
  GMPArea.CameraPosition.Y:=Y;
  GMPArea.CameraPosition.Z:=Z;
  GMPArea.ViewPoint.X:=X;
  GMPArea.ViewPoint.Y:=Y;
  GMPArea.ViewPoint.Z:=Z;
  TA_RenderAll:=TRUE;
  End;  {TTomasAnimationForm.SetCameraPosition}

Procedure TTomasAnimationForm.GetCameraPosition(Var X,Y,Z: Single);
  Begin
  X:=GMPArea.CameraPosition.X;
  Y:=GMPArea.CameraPosition.Y;
  Z:=GMPArea.CameraPosition.Z;
  End;  {TTomasAnimationForm.GetCameraPosition}

Procedure TTomasAnimationForm.SetCameraRotation(X,Y,Z: Single);
  Begin
  GMPArea.CameraRotation.X:=X * Pi /180.0;
  GMPArea.CameraRotation.Y:=Y * Pi /180.0;
  GMPArea.CameraRotation.Z:=Z * Pi /180.0;
  GMPArea.SetCameraSinCos;
  TA_RenderAll:=TRUE;
  End; {TTomasAnimationForm.SetCameraRotation}

procedure TTomasAnimationForm.getCameraRotation(Var X,Y,Z: Single);
  Begin
  X:=GMPArea.CameraRotation.X * 180.0 / Pi;
  Y:=GMPArea.CameraRotation.Y * 180.0 / Pi;
  Z:=GMPArea.CameraRotation.Z * 180.0 / Pi;
  End;  {TTomasAnimationForm.GetCameraRotation}

Procedure TTomasAnimationForm.SetLightSourcePosition(X,Y,Z: Single);
  Var
    P1,P2: TPoint3D;
  Begin
  P1.X:=X;
  P1.Y:=Y;
  P1.Z:=Z;
  P2.X:=0;
  P2.Y:=0;
  P2.Z:=0;
  GMPArea.SetLightSourcePosition(P1,P2);
  TA_RenderAll:=TRUE;
  End;   {TTomasAnimationForm.SetLightSourcePosition}

Procedure TTomasAnimationForm.SetBackGround(BackGround: TPicture);
  Begin
  GMPArea.LoadBackGround(BackGround);
  TA_RenderAll:=TRUE;
  End;  {TTomasAnimationForm.SetBackGround}

Procedure TTomasAnimationForm.SetBackGroundColor(BGColor: TColor);
  Begin
  GMPArea.SetBackColor(BGColor);
  TA_BGColor:=BGColor;
  If BGColor = clBlack Then
    Begin
    Canvas.Font.Color:=clWhite;
    End
  Else
    Begin
    Canvas.Font.Color:=clBlack;
    End;
  TA_RenderAll:=TRUE;
  End;



Constructor TCAppearance.Create;
  Begin
  Inherited Create;
  New(TA_ObjectLink);
  TA_ObjectLink^.TCA_ID:=1;
  TA_ObjectLink^.TCA_RenderMode:=TRenderMode(rmWireFrame);
  TA_ObjectLink^.TCA_Color:=clBlack;
  TA_ObjectLink^.TCA_Next:=TA_ObjectLink;
  TA_ObjectLink^.TCA_Pred:=TA_ObjectLink;
  TA_ObjectLink^.TCA_Object.NumberPolys:=0;
  TA_Position.X:=0;
  TA_Position.Y:=0;
  TA_Position.Z:=0;
  TA_Visible:=FALSE;
  TA_Parent:=Nil;
  TA_ToRender:=TRUE;
  TA_AngleX:=0;
  TA_AngleY:=0;
  TA_AngleZ:=0;
  New(TA_Link);
  TA_Link^.TL_Comp:=Self;
  TA_Link^.TL_Next:=TomasAnimationForm.TA_Comps;
  TA_Link^.TL_Pred:=TomasAnimationForm.TA_Comps^.TL_Pred;
  TomasAnimationForm.TA_Comps^.TL_Pred:=TA_Link;
  TA_Link^.TL_Pred^.TL_Next:=TA_Link;
  New(TA_Dependents);
  TA_Dependents^.TL_Comp:=Nil;
  TA_Dependents^.TL_Next:=TA_Dependents;
  TA_Dependents^.TL_Pred:=TA_Dependents;
  End;  {TCAppearance.Create}

Destructor TCAppearance.Destroy;
  var
    TCALink: PCA_Object;
  Begin
  TA_ObjectLink^.TCA_Pred^.TCA_Next:=Nil;
  TCALink:=TA_ObjectLink;
  While TCALink <> Nil Do
    Begin
    TA_ObjectLink:=TCALink^.TCA_Next;
    Dispose(TCALink);
    TCAlink:=TA_ObjectLink;
    End;
  TA_Link^.TL_Pred^.TL_Next:=TA_Link^.TL_Next;
  TA_Link^.TL_Next^.TL_Pred:=TA_Link^.TL_Pred;
  Dispose(TA_Link);
  Inherited Destroy;
  End;  {TCAppearance.Done}

Procedure TCAppearance.LoadFromFile(FileName: String);
  Var
    S: String;
    TCALink: PCA_Object;
  Begin
  TA_ToRender:=TRUE;
  New(TCALink);
  TCALink^.TCA_ID:=1;
  TCALink^.TCA_Next:=TA_ObjectLink;
  TCALink^.TCA_Pred:=TA_ObjectLink^.TCA_Pred;
  TCALink^.TCA_Pred^.TCA_Next:=TCALink;
  TA_ObjectLink^.TCA_Pred:=TCALink;
  TCALink^.TCA_RenderMode:=TrenderMode(rmSolidShade);
  TCALink^.TCA_Color:=clRed;
  TCALink^.TCA_XOff:=0;
  TCALink^.TCA_YOff:=0;
  TCALink^.TCA_ZOff:=0;
  If TomasAnimationForm.GMPArea.LoadGeoObject(TCALink^.TCA_Object,FileName) < 0 Then
    Begin
    Str(TNow:12:2,S);
    MessageDlg(S + ' Error(LoadFromFile): invalid or non-existing geo-file',
     mtError,[mbOK],0);
    RunError(255);
    End;

  End;  {TCAppearance.LoadFromFile}


Procedure TCAppearance.Show(X,Y,Z,Angle: Single);
  Var
    DepLink: PCompLink;
    TCALink: PCA_Object;

  procedure GoSetPosition;
    Begin
    TA_Position.X:=X;
    TA_Position.Y:=Y;
    TA_Position.Z:=Z;
    TCALink:=TA_ObjectLink^.TCA_Next;
    While TCALink <> TA_ObjectLink Do
      Begin
      TCALink^.TCA_Object.World.X:=X{ + TCALink^.TCA_XOff};
      TCALink^.TCA_Object.World.Y:=Y{ + TCALink^.TCA_YOff};
      TCALink^.TCA_Object.World.Z:=Z{ + TCALink^.TCA_ZOff};
      TCALink:=TCALink^.TCA_Next;
      End;
    DepLink:=TA_Dependents^.TL_Next;
    While DepLink<>TA_Dependents Do
      Begin
      DepLink^.TL_Comp.TA_ToRender:=TRUE;
      DepLink:=DepLink^.TL_Next;
      End;
    End;

  Begin
  TA_Visible:=TRUE;
  TA_ToRender:=TRUE;
  If Angle <> 0 Then
    Begin
    TA_AngleX:=TA_AngleX + X * Angle;
    TA_AngleY:=TA_AngleY + Y * Angle;
    TA_AngleZ:=TA_AngleZ + Z * Angle;
    TCALink:=TA_ObjectLink^.TCA_Next;
    While TCALink <> TA_ObjectLink Do
      Begin
      With TCALink^ Do
        Begin
        TomasAnimationForm.GMPArea.Rotate(X,Y,Z,Angle*Pi/180.0,TCA_Object);
        End;
      TCALink:=TCALink^.TCA_Next;
      End;
    DepLink:=TA_Dependents^.TL_Next;
    While DepLink<>TA_Dependents Do
      Begin
      DepLink^.TL_Comp.Show(X,Y,Z,Angle);
      DepLink^.TL_Comp.TA_ToRender:=TRUE;
      DepLink:=DepLink^.TL_Next;
      End;
    End
  Else
    Begin
    GoSetPosition;
    End;
  End;  {TCAppearance.Show}

Procedure TCAppearance.Hide;
  Begin
  TA_Visible:=FALSE;
  End;  {TCAppearance.Hide}


Procedure TCAppearance.GetPosition(Var X,Y,Z: Single);
  Begin
  If TA_Parent = Nil Then
    begin
    X:=TA_Position.X;
    Y:=TA_Position.Y;
    Z:=TA_Position.Z;
    End
  Else
    Begin
    TA_Parent.GetPosition(X,Y,Z);
    X:=X + TA_Position.X;
    Y:=Y + TA_Position.Y;
    Z:=Z + TA_Position.Z;
    End;
  End;  {TCAppearance.GetPosition}

Function TCAppearance.GetAngleX: Single;
  Begin
  Result:=TA_AngleX;
  End;

Function TCAppearance.GetAngleY: Single;
  Begin
  Result:=TA_AngleY;
  End;

Function TCAppearance.GetAngleZ: Single;
  Begin
  Result:=TA_AngleZ;
  End;

Procedure TCAppearance.SetRenderMode(rm: TRenderMode);
  Var
    TCALink: PCA_Object;
  Begin
  TA_ToRender:=TRUE;
  TCALink:=TA_ObjectLink^.TCA_Next;
  While TCALink <> TA_ObjectLink Do
    Begin
    TCALink^.TCA_RenderMode:=rm;
    TCALink:=TCALink^.TCA_Next;
    End;
  End;  {}

Procedure TCAppearance.SetColor(Color: TColor);
  Var
    TCALink: PCA_Object;
  Begin
  TA_ToRender:=TRUE;
  TCALink:=TA_ObjectLink^.TCA_Next;
  While TCALink <> TA_ObjectLink Do
    Begin
    TCALink^.TCA_Color:=Color;
    TCALink:=TCALink^.TCA_Next;
    End;
  End;  {}

Procedure TCAppearance.AddBoxObject(X,Y,Z: Single; Rm: TrenderMode;
 Color: TColor; Rx,Ry,Rz: Integer; Angle: Single; Dx,Dy,Dz: Single);
  Var
    PointArray: Array[0..7] Of TPoint3D;
    i: Integer;
    TCALink: PCA_Object;
  Const
    PlArray: Array[0..23] Of Integer =
     (3,1,2,0,0,2,6,4,6,2,1,5,7,5,1,3,7,3,0,4,4,6,5,7);

  Begin
  TA_ToRender:=TRUE;
  For i:=0 To 3 Do
    Begin
    PointArray[i].X:=-0.5 * X;
    PointArray[i+4].X:=0.5 * X;
    PointArray[2 * i].Z:=-0.5 * Z;
    PointArray[2 * i + 1].Z:=0.5 * z;
    End;
  PointArray[0].y:=-0.5 * Y;
  PointArray[3].y:=PointArray[0].y ;
  PointArray[4].y:=PointArray[0].y ;
  PointArray[7].y:=PointArray[0].y ;
  PointArray[1].y:=0.5 * Y ;
  PointArray[2].y:=PointArray[1].y ;
  PointArray[5].y:=PointArray[1].y ;
  PointArray[6].y:=PointArray[1].y ;
  New(TCALink);
  TCALink^.TCA_ID:=1;
  TCALink^.TCA_Next:=TA_ObjectLink;
  TCALink^.TCA_Pred:=TA_ObjectLink^.TCA_Pred;
  TCALink^.TCA_Pred^.TCA_Next:=TCALink;
  TA_ObjectLink^.TCA_Pred:=TCALink;
  TCALink^.TCA_RenderMode:=Rm;
  TCALink^.TCA_Color:=Color;
  TCALink^.TCA_XOff:=Dx;
  TCALink^.TCA_YOff:=Dy;
  TCALink^.TCA_ZOff:=Dz;
  For i:=0 To 5 Do
    Begin
    With TCALink^.TCA_Object.PolyStore[i] Do
      Begin
      Point[0] := PointArray[PlArray[4 * i]];
      Point[1] := PointArray[PlArray[4 * i + 1]];
      Point[2] := PointArray[PlArray[4 * i + 2]];
      Point[3] := PointArray[PlArray[4 * i + 3]];
      NumberPoints := 4;
      End;
    End;
  TCALink^.TCA_Object.NumberPolys:=6;
  TomasAnimationForm.GMPArea.Rotate(Rx,Ry,Rz,
   Angle*Pi/180.0,TCALink^.TCA_Object);
  For i:=0 to 5 Do
    Begin
    With TCALink^.TCA_Object.PolyStore[i] Do
      Begin
      Point[0].X:=Point[0].X + Dx;
      Point[1].X:=Point[1].X + Dx;
      Point[2].X:=Point[2].X + Dx;
      Point[3].X:=Point[3].X + Dx;
      Point[0].Y:=Point[0].Y + Dy;
      Point[1].Y:=Point[1].Y + Dy;
      Point[2].Y:=Point[2].Y + Dy;
      Point[3].Y:=Point[3].Y + Dy;
      Point[0].Z:=Point[0].Z + Dz;
      Point[1].Z:=Point[1].Z + Dz;
      Point[2].Z:=Point[2].Z + Dz;
      Point[3].Z:=Point[3].Z + Dz;
      End;
    End;
  end;  {TCAppearance.AddBoxObject}

Procedure TCAppearance.AddCircleObject(R: Single; Rm: TRenderMode;
 Color: TColor; Rx,Ry,Rz: Integer; Angle: Single; Dx,Dy,Dz: Single);
  Var
    PointArray: Array[0..16] Of TPoint3D;  {0 is centre, 1..16 is on circle}
    i: Integer;
    Da: Single;
    TCALink: PCA_Object;
  Begin
  TA_ToRender:=TRUE;
  PointArray[0].X:=0;
  PointArray[0].Y:=0;
  PointArray[0].Z:=0;
  Da:=0;
  For i:=1 To 16 Do
    Begin
    PointArray[i].X:=0;
    PointArray[i].Y:=R * sin(Da * Pi/180.0);
    PointArray[i].Z:=R * cos(Da * Pi/180.0);
    Da:=Da + 22.5;
    End;
  New(TCALink);
  TCALink^.TCA_ID:=1;
  TCALink^.TCA_Next:=TA_ObjectLink;
  TCALink^.TCA_Pred:=TA_ObjectLink^.TCA_Pred;
  TCALink^.TCA_Pred^.TCA_Next:=TCALink;
  TA_ObjectLink^.TCA_Pred:=TCALink;
  TCALink^.TCA_RenderMode:=Rm;
  TCALink^.TCA_Color:=Color;
  TCALink^.TCA_XOff:=Dx;
  TCALink^.TCA_YOff:=Dy;
  TCALink^.TCA_ZOff:=Dz;
  For i:=0 To 15 Do
    Begin
    With TCALink^.TCA_Object.PolyStore[i] Do
      Begin
      Point[0] := PointArray[0];
      Point[1] := PointArray[i + 1];
      If i = 15 Then
        begin
        Point[2]:=PointArray[1];
        End
      Else
        Begin
        Point[2] := PointArray[i + 2];
        End;
      NumberPoints := 3;
      End;
    End;
  TCALink^.TCA_Object.NumberPolys:=16;
  TomasAnimationForm.GMPArea.Rotate(Rx,Ry,Rz,
   Angle*Pi/180.0,TCALink^.TCA_Object);
  For i:=0 to 15 Do
    Begin
    With TCALink^.TCA_Object.PolyStore[i] Do
      Begin
      Point[0].X:=Point[0].X + Dx;
      Point[1].X:=Point[1].X + Dx;
      Point[2].X:=Point[2].X + Dx;
      Point[0].Y:=Point[0].Y + Dy;
      Point[1].Y:=Point[1].Y + Dy;
      Point[2].Y:=Point[2].Y + Dy;
      Point[0].Z:=Point[0].Z + Dz;
      Point[1].Z:=Point[1].Z + Dz;
      Point[2].Z:=Point[2].Z + Dz;
      End;
    End;
  End;   {TCAppearance.AddCircleObject}

Procedure TCAppearance.AddCylinderObject(H,R: Single; Rm: TRenderMode;
 Color: TColor; Rx,Ry,Rz: Integer; Angle: Single; Dx,Dy,Dz: Single);
  Var
    PointArray: Array[0..49] Of TPoint3D;
    i: Integer;
    TCALink: PCA_Object;
    H5: Single;
  Begin
  TA_ToRender:=TRUE;
  H5:=0.5 * H;
  For i:=0 To 23 Do
    Begin
    PointArray[i].Z:=H5;
    PointArray[24 + i].Z:=-H5;
    End;
  PointArray[0].X:=R;
  PointArray[0].Y:=0;
  PointArray[6].X:=0;
  PointArray[6].Y:=R;
  PointArray[12].X:=-R;
  PointArray[12].Y:=0;
  PointArray[18].X:=0;
  PointArray[18].Y:=-R;
  PointArray[24].X:=R;
  PointArray[24].Y:=0;
  PointArray[30].X:=0;
  PointArray[30].Y:=R;
  PointArray[36].X:=-R;
  PointArray[36].Y:=0;
  PointArray[42].X:=0;
  PointArray[42].Y:=-R;
  PoinTArray[48].X:=0;
  PointArray[48].Y:=0;
  PointArray[48].Z:=H5;
  PoinTArray[49].X:=0;
  PointArray[49].Y:=0;
  PointArray[49].Z:=-H5;
  For i:=1 To 5 Do
    Begin
    PointArray[i].X:=R * Cos(15 * i * Pi/180.0);
    PointArray[12 - i].X := - PointArray[i].X;
    PointArray[12 + i].X := - PointArray[i].X;
    PointArray[24 - i].X := PointArray[i].X;

    PointArray[24 + i].X := PointArray[i].X;
    PointArray[36 - i].X := - PointArray[i].X;
    PointArray[36 + i].X := - PointArray[i].X;
    PointArray[48 - i].X := PointArray[i].X;

    PointArray[i].Y:=R * Sin(15 * i * Pi/180.0);
    PointArray[12 - i].Y := PointArray[i].Y;
    PointArray[12 + i].Y := - PointArray[i].Y;
    PointArray[24 - i].Y := - PointArray[i].Y;

    PointArray[24 + i].Y := PointArray[i].Y;
    PointArray[36 - i].Y := PointArray[i].Y;
    PointArray[36 + i].Y := - PointArray[i].Y;
    PointArray[48 - i].Y := - PointArray[i].Y;
    End;
  New(TCALink);
  TCALink^.TCA_ID:=1;
  TCALink^.TCA_Next:=TA_ObjectLink;
  TCALink^.TCA_Pred:=TA_ObjectLink^.TCA_Pred;
  TCALink^.TCA_Pred^.TCA_Next:=TCALink;
  TA_ObjectLink^.TCA_Pred:=TCALink;
  TCALink^.TCA_RenderMode:=Rm;
  TCALink^.TCA_Color:=Color;
  For i:=0 To 23 Do
    Begin
    With TCALink^.TCA_Object.PolyStore[i] Do
      Begin
      Point[0] := PointArray[i];
      Point[1] := PointArray[(i + 1) mod 24];
      Point[2] := PointArray[24 + (i + 1) mod 24];
      Point[3] := PointArray[24 + i];
      NumberPoints := 4;
      End;
    With TCALink^.TCA_Object.PolyStore[24 + i] Do
      Begin
      Point[0]:= PointArray[i];
      Point[1]:= PointArray[48];
      Point[2]:= PointArray[(i + 1) mod 24];
      NumberPoints := 3;
      End;
    With TCALink^.TCA_Object.PolyStore[48 + i] Do
      Begin
      Point[0]:= PointArray[24 + i];
      Point[1]:= PointArray[49];
      Point[2]:= PointArray[24 + (i + 1) mod 24];
      NumberPoints := 3;
      End;
    End;
  TCALink^.TCA_Object.NumberPolys:=72;
  TomasAnimationForm.GMPArea.Rotate(Rx,Ry,Rz,
   Angle*Pi/180.0,TCALink^.TCA_Object);
  For i:=0 To 71 Do
    Begin
    With TCALink^.TCA_Object.PolyStore[i] Do
      Begin
      Point[0].X:=Point[0].X + Dx;
      Point[1].X:=Point[1].X + Dx;
      Point[2].X:=Point[2].X + Dx;
      Point[0].Y:=Point[0].Y + Dy;
      Point[1].Y:=Point[1].Y + Dy;
      Point[2].Y:=Point[2].Y + Dy;

      Point[0].Z:=Point[0].Z + Dz;
      Point[1].Z:=Point[1].Z + Dz;
      Point[2].Z:=Point[2].Z + Dz;
      If NumberPoints = 4 Then
        Begin
        Point[3].X:=Point[3].X + Dx;
        Point[3].Y:=Point[3].Y + Dy;
        Point[3].Z:=Point[3].Z + Dz;
        End;
      End;
    End;
  End;   {TCAppearance.AddCylinderObject}

Procedure TCAppearance.Add(TCApp: TCAppearance; X,Y,Z: Single);
  Var
    DepLink: PCompLink;
    DepLink2: PCompLink;
    Found: Boolean;
  Begin
  TA_ToRender:=TRUE;
  DepLink2:=Nil;
  DepLink:=TomasAnimationForm.TA_Comps^.TL_Next;
  While (DepLink<>TomasAnimationForm.TA_Comps)
   And (DepLink^.TL_Comp<>TCApp) Do
    Begin
    DepLink:=DepLink^.TL_Next;
    End;
  If DepLink=TomasAnimationForm.TA_Comps Then
    Begin   {search if appearance already combined}
    Found:=FALSE;
    DepLink:=TomasAnimationForm.TA_Comps^.TL_Next;
    While (DepLink<>TomasAnimationForm.TA_Comps) And NOT(Found) Do
      Begin
      DepLink2:=DepLink^.TL_Comp.TA_Dependents^.TL_Next;
      While (DepLink2<>DepLink^.TL_Comp.TA_Dependents) And NOT(Found) Do
        Begin
        If DepLink2^.TL_Comp = TCApp Then
          Begin
          Found:=TRUE;
          End
        Else
          Begin
          DepLink2:=DepLink2^.TL_Next;
          End;
        End;
      If NOT(Found) Then
        Begin
        DepLink:=DepLink^.TL_Next;
        End;
      End;
    If Found Then
      Begin
      DepLink2^.TL_Next^.TL_Pred:=DepLink2^.TL_Pred;
      DepLink2^.TL_Pred^.TL_Next:=DepLink2^.TL_Next;
      DepLink2^.TL_Next:=TomasAnimationForm.TA_Comps;
      DepLink2^.TL_Pred:=TomasAnimationForm.TA_Comps^.TL_Pred;
      TomasAnimationForm.TA_Comps^.TL_Pred:=DepLink2;
      DepLink2^.TL_Pred^.TL_Next:=DepLink2;
      DepLink:=DepLink2;
      End;
    End;
  If DepLink<>TomasAnimationForm.TA_Comps Then
    Begin
    DepLink^.TL_Next^.TL_Pred:=DepLink^.TL_Pred;
    DepLink^.TL_Pred^.TL_Next:=DepLink^.TL_Next;
    DepLink^.TL_Next:=TA_Dependents;
    DepLink^.TL_Pred:=TA_Dependents^.TL_Pred;
    TA_Dependents^.TL_Pred:=DepLink;
    DepLink^.TL_Pred^.TL_Next:=DepLink;
    DepLink^.TL_Comp.TA_Position.X:=X;
    DepLink^.TL_Comp.TA_Position.Y:=Y;
    DepLink^.TL_Comp.TA_Position.Z:=Z;
    DepLink^.TL_Comp.TA_Parent:=Self;
    End
  Else
    Begin
    MessageDlg('Appearance to combine not found',mtError,[mbOK],0);
    RunError(255);
    End;
  End;   {TCAppearance.Add}

Procedure TCAppearance.UnTie;
  Var
    DepLink: PCompLink;
    DepLink2: PCompLink;
    Found: Boolean;
  Begin
  TA_ToRender:=TRUE;
  Found:=FALSE;
  DepLink2:=Nil;
  DepLink:=TomasAnimationForm.TA_Comps^.TL_Next;
  While (DepLink<>TomasAnimationForm.TA_Comps) And NOT(Found) Do
    Begin
    DepLink2:=DepLink^.TL_Comp.TA_Dependents^.TL_Next;
    While (DepLink2<>DepLink^.TL_Comp.TA_Dependents) And NOT(Found) Do
      Begin
      If DepLink2^.TL_Comp = Self Then
        Begin
        Found:=TRUE;
        End
      Else
        Begin
        DepLink2:=DepLink2^.TL_Next;
        End;
      End;
    If NOT(Found) Then
      Begin
      DepLink:=DepLink^.TL_Next;
      End;
    End;
  If Found Then
    Begin
    DepLink2^.TL_Next^.TL_Pred:=DepLink2^.TL_Pred;
    DepLink2^.TL_Pred^.TL_Next:=DepLink2^.TL_Next;
    DepLink2^.TL_Next:=TomasAnimationForm.TA_Comps;
    DepLink2^.TL_Pred:=TomasAnimationForm.TA_Comps^.TL_Pred;
    TomasAnimationForm.TA_Comps^.TL_Pred:=DepLink2;
    DepLink2^.TL_Pred^.TL_Next:=DepLink2;
    TA_Position.X:=TA_Position.X + TA_Parent.TA_Position.X;
    TA_Position.Y:=TA_Position.Y + TA_Parent.TA_Position.Y;
    TA_Position.Z:=TA_Position.Z + TA_Parent.TA_Position.Z;
    TA_Parent:=Nil;
    End
  Else
    Begin
    MessageDlg('Appearance to free not found',mtError,[mbOK],0);
    RunError(255);
    End;
  End;    {TCAppearance.UnTie}

Constructor TCObject.Create;
  Begin
  Inherited Create;
  TO_RenderMode:=TRenderMode(rmWireFrame);
  TO_Color:=clYellow;
  TO_Position.X:=0;
  TO_Position.Y:=0;
  TO_Position.Z:=0;
  TO_Angle:=0;
  TO_ToRender:=TRUE;
  New(TO_Link);
  TO_Link^.TO_Object:=Self;
  TO_Link^.TO_ID:=0;
  TO_Link^.TO_Next:=TomasAnimationForm.TA_Objects;
  TO_Link^.TO_Pred:=TomasAnimationForm.TA_Objects^.TO_Pred;
  TomasAnimationForm.TA_Objects^.TO_Pred:=TO_Link;
  TO_Link^.TO_Pred^.TO_Next:=TO_Link;
  End;  {TCObject.Create}

Destructor TCObject.Destroy;
  Begin
  TO_Link^.TO_Pred^.TO_Next:=TO_Link^.TO_Next;
  TO_Link^.TO_Next^.TO_Pred:=TO_Link^.TO_Pred;
  Dispose(TO_Link);
  Inherited Destroy;
  End;  {TCObject.Destroy}

Procedure TCObject.ReadAppearanceFromFile(FileName: String);
  Var
    S: String;
  Begin
  If TomasAnimationForm.GMPArea.LoadGeoObject(TO_Object,FileName) < 0 Then
    Begin
    Str(TNow:12:2,S);
    MessageDlg(S + ' Error(LoadFromFile): invalid or non-existing geo-file',
     mtError,[mbOK],0);
    RunError(255);
    End
  Else
    begin
    TO_ToRender:=TRUE;
    End;
  End;  {TCObject.LoadFromFile}

Procedure TCObject.Show(X,Y,Z,Angle: Single);
  Begin
  TO_Position.X:=X;
  TO_Position.Y:=Y;
  TO_Position.Z:=Z;
  TO_Object.World.X:=X;
  TO_Object.World.Y:=Y;
  TO_Object.World.Z:=Z;
  TO_Angle:=Angle;
  TO_ToRender:=TRUE;
  End;  {TCObject.Show}

Procedure TCObject.SetRenderMode(Rm: TRenderMode);
  Begin
  TO_RenderMode:=Rm;
  TO_ToRender:=TRUE;
  End;  {TCObject.SetRenderMode}


Function ZYSort(Item1,Item2: Pointer): Integer;
  Var
    R: Single;
    R1: Single;
    R2: Single;
    Point1: TPoint3D;
    Point2: TPoint3D;
    X,Y,Z: Single;
  Begin
  TomasAnimationForm.GetCameraPosition(X,Y,Z);
  If PObjLink(Item1)^.TO_ID = 0 Then
    Begin
    Point1.Z:=PObjLink(Item1)^.TO_Object.TO_Object.World.Z - Z;
    Point1.Y:=PObjLink(Item1)^.TO_Object.TO_Object.World.Y - Y;
    Point1.X:=PObjLink(Item1)^.TO_Object.TO_Object.World.X - X;
    End
  Else
    Begin
    Point1.Z:=PCA_Object(Item1)^.TCA_Center.Z - Z;
    Point1.Y:=PCA_Object(Item1)^.TCA_Center.Y - Y;
    Point1.X:=PCA_Object(Item1)^.TCA_Center.X - X;
    End;
  If PObjLink(Item2)^.TO_ID = 0 Then
    Begin
    Point2.Z:=PObjLink(Item2)^.TO_Object.TO_Object.World.Z - Z;
    Point2.Y:=PObjLink(Item2)^.TO_Object.TO_Object.World.Y - Y;
    Point2.X:=PObjLink(Item2)^.TO_Object.TO_Object.World.X - X;
    End
  Else
    Begin
    Point2.Z:=PCA_Object(Item2)^.TCA_Center.Z - Z;
    Point2.Y:=PCA_Object(Item2)^.TCA_Center.Y - Y;
    Point2.X:=PCA_Object(Item2)^.TCA_Center.X - X;
    End;
  R1:=Sqrt(Point1.Z * Point1.Z + Point1.Y * Point1.Y + Point1.X * Point1.X);
  R2:=Sqrt(Point2.Z * Point2.Z + Point2.Y * Point2.Y + Point2.X * Point2.X);
  R:=R2 - R1;
  Result:=Round(100000*R);
  End;  {ZYSort}

Procedure TTomasAnimationForm.SortOnZOrder;
  Var
    NxtLink: PCompLink;
    DepLink: PCompLink;
    ObjLink: PObjLink;
    TCALink: PCA_Object;
  Begin
  ObjLink:=TA_Objects^.TO_Next;
  While ObjLink <> TA_Objects Do
    Begin
    If ObjLink^.TO_Object.TO_ToRender Or TA_RenderAll Then
      Begin
      GMPArea.PrepareRender(ObjLink^.TO_Object.TO_Object);
      ObjLink^.TO_Object.TO_ToRender:=FALSE;
      End;
    TA_SortList.Add(ObjLink);
    ObjLink:=ObjLink^.TO_Next;
    End;
  NxtLink:=TA_Comps^.TL_Next;
  While NxtLink <> TA_Comps Do
    Begin
    If NxtLink^.TL_Comp.TA_Visible Then
      Begin
      With NxtLink^.TL_Comp Do
        Begin
        TCALink:=TA_ObjectLink^.TCA_Next;
        While TCALink<>TA_ObjectLink Do
          Begin
          If TA_ToRender Or TA_RenderAll Then
            Begin
            With TCALink^ Do
              Begin
              TCA_Object.World.X:=NxtLink^.TL_Comp.TA_Position.X;
              TCA_Object.World.Y:=NxtLink^.TL_Comp.TA_Position.Y;
              TCA_Object.World.Z:=NxtLink^.TL_Comp.TA_Position.Z;
              End;
            GMPArea.ChangeObjectColor(TCALink^.TCA_Object,TCALink^.TCA_Color);
            GMPArea.RenderMode:=TCALink^.TCA_RenderMode;
            GMPArea.PrepareRender(TCALink^.TCA_Object);
            End;
         { GMPArea.GetObjectCenter(TCALink^.TCA_Object,TCALink^.TCA_Center);}
          TA_SortList.Add(TCALink);
          TCALink:=TCALink^.TCA_Next;
          End;
        End;
      DepLink:=NxtLink^.TL_Comp.TA_Dependents^.TL_Next;
      While DepLink<>NxtLink^.TL_Comp.TA_Dependents Do
        Begin
        With DepLink^.TL_Comp Do
          Begin
          TCALink:=TA_ObjectLink^.TCA_Next;
          While TCALink<>TA_ObjectLink Do
            Begin
            If TA_ToRender Or TA_RenderAll Then
              Begin
              With TCALink^ Do
                Begin
                TCA_Object.World.X:=NxtLink^.TL_Comp.TA_Position.X + TA_Position.X;
                TCA_Object.World.Y:=NxtLink^.TL_Comp.TA_Position.Y + TA_Position.Y;
                TCA_Object.World.Z:=NxtLink^.TL_Comp.TA_Position.Z + TA_Position.Z;
                End;
              GMPArea.ChangeObjectColor(TCALink^.TCA_Object,TCALink^.TCA_Color);
              GMPArea.RenderMode:=TCALink^.TCA_RenderMode;
              GMPArea.PrepareRender(TCALink^.TCA_Object);
              End;
            {GMPArea.GetObjectCenter(TCALink^.TCA_Object,TCALink^.TCA_Center);}
            TA_SortList.Add(TCALink);
            TCALink:=TCALink^.TCA_Next;
            End;
          TA_ToRender:=FALSE;
          End;
        DepLink:=DepLink^.TL_Next;
        End;
      End;
    NxtLink^.TL_Comp.TA_ToRender:=FALSE;
    NxtLink:=NxtLink^.TL_Next;
    End;
  {TA_SortList.Sort(ZYSort);}
  End;   {TTomasAnimationForm.SortOnZOrder}


procedure TTomasAnimationForm.Refresh(Sender: TObject);
  var
    SortItem: Pointer;
begin
SortOnZOrder;
GMPArea.ClearBackPage;
While TA_SortList.Count > 0 Do
  Begin
  SortItem:=TA_SortList.First;
  If PObjLink(SortItem)^.TO_ID=0 Then {object}
    Begin
    With PObjLink(SortItem)^.TO_Object Do
      Begin
      GMPArea.ChangeObjectColor(TO_Object,clGray);
      GMPArea.RenderMode:=TO_RenderMode;
      GMPArea.RenderNow(TO_Object,TRUE);
      End;
    End
  Else
    Begin
    With PCA_Object(SortItem)^ Do
      begin
      GMPArea.ChangeObjectColor(TCA_Object,TCA_Color);
      GMPArea.RenderMode:=TCA_RenderMode;
      GMPArea.RenderNow(TCA_Object,TRUE);
      End;
    End;
  TA_SortList.Remove(SortItem);
  End;
GMPArea.FlipBackPage;
TA_RenderAll:=FALSE;
end;

Procedure TTomasAnimationForm.DrawLine(X1,Y1,Z1,X2,Y2,Z2: Single; Color: TColor);
  Begin
  GMPArea.RenderMode:=TRenderMode(rmWireFrame);
  GMPArea.DrawLine(X1,Y1,Z1,X2,Y2,Z2,Color);
  End;  {TTomasAnimationForm.DrawLine}

procedure TTomasAnimationForm.SetSpeedPosition(X: Double);
  Var
    P: Integer;
  Begin
  If X<=0 Then
    Begin
    X:=0;
    End;
  If X <= 0.1 Then
    Begin
    P:=Round(100*X);
    End
  Else
    begin
    If X<=0.5 Then
      Begin
      P:=Round(25*X + 7.5);
      End
    Else
      Begin
      if X<=1 Then
        begin
        P:=Round(20 * X + 10);
        End
      Else
        Begin
        If X <=10 Then
          begin
          P:=Round(4* X + 26)
          End
        Else
          Begin
          P:=Round(66 + X/33);
          End;
        End;
      End;
    End;
  TrackBar1.Position:=P;
  End;

procedure TTomasAnimationForm.GoShowData(Canvas: TCanvas);
  begin
  Canvas.Font.Name := 'Arial';
  Canvas.Font.Style := [fsBold, fsItalic];
  Canvas.Brush.Style := bsClear;
  Canvas.Font.Size := 12;
  If TA_BGColor = clBlack Then
    Begin
    Canvas.Font.Color:=clWhite;
    End
  Else
    Begin
    Canvas.Font.Color:=clBlack;
    End;
  Canvas.Textout (10,10,TScreenString);
  Canvas.Brush.Style := bsSolid;
  If Assigned(PBackGround) Then
    begin
    GMPArea.BackGroundDrawing:=TRUE;
    PBackGround;
    GMPArea.BackGroundDrawing:=FALSE;
    End;
  end;

procedure TTomasAnimationForm.CheckPosition(Sender: TObject);
Var
  P: Integer;
  X: Double;
begin
P:=TrackBar1.Position;
if P <=10 Then
  begin
  X:=P/100;
  End
Else
  Begin
  If P<=20 Then
    Begin
    X:=0.1 + (P-10)/25;
    End
  Else
    Begin
    If P <=30 Then
      Begin
      X:=0.5 + (P-20)/20;
      End
    Else
      Begin
      If P <= 70 Then
        begin
        X:=1 + (P-30)/4;
        End
      Else
        Begin
        X:=10 + (P-70) * 33;
        End
      End;
    End;
  End;
SetAnimationDisplayInterval(X);
end;

procedure TTomasAnimationForm.CheckCameraMovement(Sender: TObject;
  var Key: Word; Shift: TShiftState);
begin
KeyIsDown:=TRUE;
If (Key >= VK_Left) And (Key <= VK_Down) Then
  Begin
  KeyIsDown:=TRUE;
  CurrentKeyDown:=Key;
  CurrentShiftState:=Shift;
  TrackBar1.Enabled:=FALSE;
  MoveCamera(CurrentKeyDown,CurrentShiftState,0.2);
  End;
end;

procedure TTomasAnimationForm.CheckCameraMovementEnd(Sender: TObject;
  var Key: Word; Shift: TShiftState);
begin
If KeyIsDown Then
  Begin
  KeyIsDown:=FALSE;
  TrackBar1.Enabled:=TRUE;
  End;
end;

procedure TTomasAnimationForm.OpenCameraForm(Sender: TObject);
begin
TomasForm.GoInterruptSimulation(Nil);
Application.CreateForm(TCameraForm,CameraForm);
CameraForm.Show;
CameraForm.BringToFront;
end;

Procedure TTomasAnimationForm.GoResumeSimulation;
  Begin
  TomasForm.GoResumeSimulation(Nil);
  End;

end.

