unit TomasResources;

(***********************************************************************)
(*  This software is copyleft software;                                *)
(*  You may distribute or change the software as you wish, but         *)
(*  this header must always be present.                                *)
(*  No one is allowed to claim copyright on this source code.          *)
(*  All precompiled software using this unit should be accompanied     *)
(*  by this source code.                                               *)
(*                                                                     *)
(*  Original contents on http://home.planet.nl/~veekehpm               *)
(*  Contact: Hans P.M. Veeke         H.P.M.Veeke@wbmt.tudelft.nl       *)
(***********************************************************************)

(***********************************************************************)
(* Update history                                                      *)
(* 20-06-2000: incorrect error message when resuming a claiming        *)
(*             element solved                                          *)
(***********************************************************************)

interface

uses
  Tomas,
  TomasProcess,
  Classes,
  Dialogs;

Type
  TomasSemaphore = class;
  TomasResource = class;
  SelectionFunction = Function: TomasElement;

TomasSemaphore = class
  Private
    FName: String;
    FMinimum: Double;
    FMaximum: Double;
    FCurrentLevel: Double;
    FSelection: SelectionFunction;
    FClaimQueue: TomasQueue;
    FUseQueue: TomasQueue;
    FResource: TomasResource;
    Procedure HonorClaim(C: TomasElement);
    Procedure CheckClaims;
  Published
    Constructor Create(SName: String; CMin,CMax: Double; Selection: SelectionFunction);
    Destructor Destroy; Override;
    Function Available: Double;
    Procedure Claim(X: Double);
    Procedure Release(X: Double);
    Function FirstClaimer: TomasElement;
    Function NextClaimer(C: TomasElement): TomasElement;
    Function NrOfClaimsWaiting: Word;
    Function NrOfUsers: Word;
    Procedure ChangeCapacityBounds(CMin,CMax: Double);
    Function UseLevel: Double;
    Function ClaimSizeOf(Elm: TomasElement): Double;
  End;


TomasResource = class(TomasElement)
  Private
    FSemaphore: TomasSemaphore;
    FInternal: TomasQueue;
    FOutput: TomasQueue;
    FDelivery: TList;
  Published
    Constructor Create(RName: String; CMin,CMax: Double;
     Selection: SelectionFunction; Delivery: TList);
    Destructor Destroy; Override;
    Procedure Process; Override;
    Function Available: Double;
    Procedure Claim(Item: Pointer; X,Duration: Double);
    Function FirstClaimer: Pointer;
    Function NextClaimer(P: Pointer): Pointer;
    Function NrOfClaimsWaiting: Word;
    Function NrOfUsers: Word;
    Procedure ChangeCapacityBounds(CMin,CMax: Double);
    Function UseLevel: Double;
    Function CheckReady(Item: Pointer): Double;
    Function FirstOutItem: Pointer;
    Procedure DeliverItem(Item: Pointer);
  End;

Function FIFOSelect: TomasElement;
Function LIFOSelect: TomasElement;
Function SCCSelect: TomasElement;
Function GCCSelect: TomasElement;


implementation

Type

  TClaimItem = class
    CI_Semaphore: TomasSemaphore;
    CI_Size: Double;
    End;

  ResourceItem = class(TomasElement)
    public
    RI_Item: Pointer;
    RI_Size: Double;
    RI_Duration: Double;
    RI_Ready: Double;
    End;

Var
  _CurrentSemaphore: TomasSemaphore;
  CurrentClaimElement: TomasElement;

Constructor TomasSemaphore.Create(SName: String; CMin,CMax: Double;
 Selection: SelectionFunction);
  Var
    OldTrace: Boolean;
  Begin
  Inherited Create;
  FName:=SName;
  FMinimum:=CMin;
  FMaximum:=CMax;
  FCurrentLevel:=CMin;
  FSelection:=Selection;
  FResource:=Nil;
  OldTrace:=TomasForm.Trace;
  TomasForm.SetTrace(FALSE);
  FClaimQueue:=TomasQueue.Create(Sname + '_ClmQ');
  FUseQueue:=TomasQueue.Create(SName + '_UseQ');
  TomasForm.SetTrace(OldTrace);
  End;   {TomasSemaphore.Create}

Destructor TomasSemaphore.Destroy;
  Var
    OldTrace: Boolean;
  Begin
  OldTrace:=TomasForm.Trace;
  TomasForm.SetTrace(FALSE);
  FClaimQueue.Destroy;
  FUseQueue.Destroy;
  TomasForm.SetTrace(OldTrace);
  Inherited Destroy;
  End;   {TomasSemaphore.Destroy}

Function TomasSemaphore.Available: Double;
  Begin
  Result:=FMaximum - FCurrentLevel;
  End;   {TomasSemaphore.Available}

Procedure TomasSemaphore.HonorClaim(C: TomasElement);
  Var
    NewClaim: TClaimItem;
    I: Integer;
    Found: Boolean;
    OldTrace: boolean;
  Begin
  OldTrace:=TomasForm.Trace;
  TomasForm.SetTrace(FALSE);
  ClaimRoom.Remove(C);
  FClaimQueue.Remove(C);
  I:=0;
  Found:=FALSE;
  NewClaim:=Nil;
  While (I<C.Claims.Count) And NOT(Found) Do
    Begin
    NewClaim:=C.Claims[I];
    If NewClaim.CI_Semaphore = Self Then
      Begin
      Found:=TRUE;
      End
    Else
      Begin
      Inc(I);
      End;
    End;
  If Found Then
    Begin
    C.Claims.Remove(NewClaim);
    FCurrentLevel:=FCurrentLevel + NewClaim.CI_Size;
    FUseQueue.AddToTail(C);
    NewClaim.Free;
    End;
  TomasForm.SetTrace(OldTrace);
  End;  {TomasSemaphore.HonorClaim}

Procedure TomasSemaphore.Claim(X: Double);
  Var
    NewClaim: TClaimItem;
    OldTrace: Boolean;
  Begin
  If GetSimulationStatus < S_Running Then
    Begin
    MessageDlg('Error(Claim): Semaphore ' + FName + ' claimed outside simulationrun',
     mtError,[mbOK],0);
    Runerror(255);
    End
  Else
    Begin
    If FResource = Nil Then
      Begin
      CurrentClaimElement:=CurrentElement;
      End;
    If FMaximum - FCurrentLevel < X Then
      Begin
      OldTrace:=TomasForm.Trace;
      TomasForm.SetTrace(FALSE);
      FClaimQueue.AddToTail(CurrentClaimElement);
      ClaimRoom.AddToTail(CurrentClaimElement);
      NewClaim:=TClaimItem.Create;
      NewClaim.CI_Semaphore:=Self;
      NewClaim.CI_Size:=X;
      CurrentClaimElement.Claims.Add(NewClaim);
      TomasForm.SetTrace(OldTrace);
      If FResource = Nil Then
        Begin
        Suspend;
        CurrentClaimElement:=CurrentElement;
        End
      Else
        Begin
        Exit;
        End;
      If ClaimRoom.Contains(CurrentClaimElement) Then
        Begin                              {if not it has been honored by release}
        HonorClaim(CurrentClaimElement);
        End;
      End
    Else
      Begin
      OldTrace:=TomasForm.Trace;
      TomasForm.SetTrace(FALSE);
      FCurrentLevel:=FCurrentLevel + X;
      FUseQueue.AddToTail(CurrentClaimElement);
      TomasForm.SetTrace(OldTrace);
      End;
    End;
  End;   {TomasSemaphore.Claim}

Procedure TomasSemaphore.CheckClaims;
  Var
    NextComp: TomasElement;
    ResItem: ResourceItem;
    NxtItem: ResourceItem;
    OldTrace: Boolean;
  begin
  OldTrace:=TomasForm.Trace;
  TomasForm.SetTrace(FALSE);
  _CurrentSemaphore:=Self;
  NextComp:=FSelection;
  _CurrentSemaphore:=Nil;
  While NextComp <> Nil Do
    Begin
    HonorClaim(NextComp);
    If FResource = Nil Then
      Begin
      NextComp.Resume(TNow);
      End
    Else
      Begin
      If (FResource <> CurrentElement)
       And (FResource.Status > Passive) Then
        Begin
        FResource.Stop;
        End;
      ResItem:=ResourceItem(NextComp);
      ResItem.RI_Ready:=TNow + ResItem.RI_Duration;
      NxtItem:=FResource.FInternal.FirstElement;
      While (NxtItem<>Nil) And (NxtItem.RI_Ready<=ResItem.RI_Ready) Do
        Begin
        NxtItem:=FResource.FInternal.Successor(NxtItem);
        End;
      If NxtItem = Nil Then
        Begin
        FResource.FInternal.AddToTail(ResItem);
        End
      Else
        Begin
        FResource.FInternal.AddBefore(ResItem,NxtItem);
        End;
      If FResource <> CurrentElement Then
        begin
        FResource.Start(TNow);
        End;
      End;
    _CurrentSemaphore:=Self;
    NextComp:=FSelection;
    _CurrentSemaphore:=Nil;
    End;
  TomasForm.SetTrace(OldTrace);
  End;  {TomasSemaphore.CheckClaims}

Procedure TomasSemaphore.Release(X: Double);
  Var
    OldTrace: Boolean;
  Begin
  OldTrace:=TomasForm.Trace;
  TomasForm.SetTrace(FALSE);
  FUseQueue.Remove(CurrentElement);
  FCurrentLevel:=FCurrentLevel - X;
  CheckClaims;
  TomasForm.SetTrace(OldTrace);
  End;   {TomasSemaphore.Release}

Function TomasSemaphore.FirstClaimer: TomasElement;
  Begin
  Result:=FClaimQueue.FirstElement;
  End;  {TomasSemaphore.FirstClaimer}

Function TomasSemaphore.NextClaimer(C: TomasElement): TomasElement;
  Begin
  If FClaimQueue.Contains(C) Then
    Begin
    Result:=FClaimQueue.Successor(C);
    End
  Else
    Begin
    Result:=Nil;
    End;
  End;  {TomasSemaphore.NextClaimer}

Function TomasSemaphore.NrOfClaimsWaiting: Word;
  Begin
  Result:=FClaimQueue.Length;
  End;  {TomasSemaphore.NrOfClaimsWaiting}

Function TomasSemaphore.NrOfUsers: Word;
  begin
  Result:=FUseQueue.Length;
  End;  {TomasSemaphore.NrOfUsers}

Procedure TomasSemaphore.ChangeCapacityBounds(CMin,CMax: Double);
  Begin
  FMinimum:=CMin;
  FMaximum:=CMax;
  CheckClaims;
  End;   {TomasSemaphore.ChangeCapacityBounds}

Function TomasSemaphore.UseLevel: Double;
  begin
  Result:=FCurrentLevel;
  End;  {TomasSemaphore.UseLevel}

Function TomasSemaphore.ClaimSizeOf(Elm: TomasElement): Double;
  Var
    Clm: TClaimItem;
    I: Integer;
  Begin
  I:=0;
  Result:=0;
  While I < Elm.Claims.Count Do
    Begin
    Clm:=Elm.Claims.Items[I];
    If Clm.CI_Semaphore = Self Then
      Begin
      Result:=Clm.CI_Size;
      I:=Elm.Claims.Count;
      End
    Else
      Begin
      Inc(I);
      End;
    End;
  End;

Function CurrentSemaphore: TomasSemaphore;
  Begin
  Result:=_CurrentSemaphore;
  End;  {CurrentSemaphore}

Constructor TomasResource.Create(RName: String; CMin,CMax: Double;
 Selection: SelectionFunction; Delivery: TList);
  Var
    OldTrace: Boolean;
  Begin
  Inherited Create(RName);
  OldTrace:=TomasForm.Trace;
  TomasForm.SetTrace(FALSE);
  FSemaphore:=TomasSemaphore.Create('S_' + RName,CMin,CMax,Selection);
  FSemaphore.FResource:=Self;
  FInternal:= TomasQueue.Create(RName + '_In');
  FOutput:=TomasQueue.Create(Rname+ '_Out');
  FDelivery:=Delivery;
  TomasForm.SetTrace(OldTrace);
  End; {TomasResource.Create}

Destructor TomasResource.Destroy;
  Begin
  FSemaphore.Destroy;
  Inherited Destroy;
  End; {TomasResource.Destroy}

Function TomasResource.Available: Double;
  Begin
  Result:=FSemaphore.Available;
  End; {TomasResource.Available}

Procedure TomasResource.Claim(Item: Pointer; X, Duration: Double);
  Var
    ResItem: ResourceItem;
    NxtItem: ResourceItem;
  Begin
  ResItem:=ResourceItem.Create('I_' + Name);
  ResItem.RI_Item:=Item;
  ResItem.RI_Size:=X;
  ResItem.RI_Duration:=Duration;
  ResItem.RI_Ready:=-1;
  CurrentClaimElement:=ResItem;
  FSemaphore.Claim(X);
  If NOT(ClaimRoom.Contains(ResItem)) Then
    Begin
    If Status > Passive Then
      Begin
      Stop;
      End;
    ResItem.RI_Ready:=TNow + ResItem.RI_Duration;
    NxtItem:=FInternal.FirstElement;
    While (NxtItem<>Nil) And (NxtItem.RI_Ready<=ResItem.RI_Ready) Do
      Begin
      NxtItem:=FInternal.Successor(NxtItem);
      End;
    If NxtItem = Nil Then
      Begin
      FInternal.AddToTail(ResItem);
      End
    Else
      Begin
      FInternal.AddBefore(ResItem,NxtItem);
      End;
    If Status = Passive Then
      Begin
      Resume(TNow);
      End
    Else
      Begin
      Start(TNow);
      End;
    End;
  End; {TomasResource.Claim}

Function TomasResource.FirstClaimer: Pointer;
  Var
    ResItem: ResourceItem;
  Begin
  ResItem:=FSemaphore.FClaimQueue.FirstElement;
  If ResItem = Nil Then
    begin
    Result:=Nil;
    End
  Else
    Begin
    Result:=ResItem.RI_Item;
    End;
  End;  {TomasResource.FirstClaimer}

Function TomasResource.NextClaimer(P: Pointer): Pointer;
  Var
    ResItem: ResourceItem;
  Begin
  {eerst item zoeken met item P}
  ResItem:=FSemaphore.FClaimQueue.FirstElement;
  While (ResItem <> Nil) And (ResItem.RI_Item<>P) Do
    Begin
    ResItem:=FSemaphore.FClaimQueue.Successor(ResItem);
    End;
  If ResItem <> Nil Then
    Begin
    ResItem:=FSemaphore.FClaimQueue.Successor(ResItem);
    End;
  If ResItem=Nil Then
    Begin
    Result:=Nil;
    End
  Else
    Begin
    Result:=ResItem.RI_Item;
    End;
  End;   {TomasResource.NextClaimer}

Function TomasResource.NrOfClaimsWaiting: Word;
  Begin
  Result:=FSemaphore.FClaimQueue.Length;
  End;   {TomasResource.NrOfClaimsWaiting}

Function TomasResource.NrOfUsers: Word;
  begin
  Result:=Finternal.Length;
  End;  {TomasResource.NrOfUsers}

Procedure TomasResource.ChangeCapacityBounds(CMin,CMax: Double);
  Begin
  FSemaphore.ChangeCapacityBounds(CMin,Cmax);
  End;   {TomasResource.ChangeCapacityBounds}

Function TomasResource.UseLevel: Double;
  Begin
  Result:=FSemaphore.UseLevel;
  End;   {TomasResource.UseLevel}

Function TomasResource.CheckReady(Item: Pointer): Double;
  Var
    ResItem: ResourceItem;
  Begin
  ResItem:=FOutput.FirstElement;
  While (ResItem <> Nil) And (ResItem.RI_Item <> Item) Do
    Begin
    ResItem:=FOutput.Successor(ResItem);
    End;
  If ResItem = Nil Then
    Begin
    Result:=-1;
    End
  Else
    Begin
    Result:=ResItem.RI_Ready;
    End;
  End;   {TomasResource.CheckReady}

Function TomasResource.FirstOutItem: Pointer;
  Var
    ResItem: ResourceItem;
  Begin
  ResItem:=FOutput.FirstElement;
  If ResItem = Nil Then
    begin
    Result:=Nil;
    End
  Else
    Begin
    Result:=ResItem.RI_Item;
    End;
  End;   {TomasResource.FirstOutItem}

Procedure TomasResource.DeliverItem(Item: Pointer);
  Var
    ResItem: ResourceItem;
  Begin
  ResItem:=FOutput.FirstElement;
  While (ResItem <> Nil) And (ResItem.RI_Item <> Item) Do
    Begin
    ResItem:=FOutput.Successor(ResItem);
    End;
  If ResItem <> Nil Then
    Begin
    FOutput.Remove(ResItem);
    ResItem.Destroy;
    End;
  End;   {Tomasresource.DeliverItem}

Procedure TomasResource.Process;
  Var
    ResItem: ResourceItem;
    OldTrace: Boolean;
  Begin
  While TRUE Do
    Begin
    While FSemaphore.FUseQueue{FInternal}.Length > 0 Do
      Begin
      ResItem:=FInternal.FirstElement;
      Hold(ResItem.RI_Ready - TNow);
      OldTrace:=TomasForm.Trace;
      TomasForm.SetTrace(FALSE);
      FInternal.Remove(ResItem);
      FSemaphore.FUseQueue.Remove(ResItem);
      TomasForm.SetTrace(OldTrace);
      FSemaphore.FCurrentLevel:=FSemaphore.FCurrentLevel - ResItem.RI_Size;
      ResItem.RI_Ready:=TNow;
      If FDelivery = Nil Then
        Begin
        FOutput.AddToTail(ResItem);
        End
      Else
        Begin
        FDelivery.Add(ResItem.RI_Item);
        ResItem.Free;
        End;
      FSemaphore.CheckClaims;
      End;
    Suspend;
    End;
  End;   {TomasResource.Process}


Function FIFOSelect: TomasElement;
  Var
    Comp: TomasElement;
    Found: Boolean;
  Begin
  With _CurrentSemaphore Do
    Begin
    Comp:=FClaimQueue.FirstElement;
    Found:=FALSE;
    While (Comp <> Nil) And NOT(Found) Do
      begin
      If ClaimSizeOf(Comp)<= FMaximum - FCurrentLevel Then
        Begin
        Found:=TRUE;
        End
      Else
        Begin
        Comp:=FClaimQueue.Successor(Comp);
        End;
      End;
    End;
  Result:=Comp;
  End;   {FIFOSelect}

Function LIFOSelect: TomasElement;
  Var
    Comp: TomasElement;
    Found: Boolean;
  Begin
  With _CurrentSemaphore Do
    Begin
    Comp:=FClaimQueue.LastElement;
    Found:=FALSE;
    While (Comp <> Nil) And NOT(Found) Do
      begin
      If ClaimSizeOf(Comp) <= FMaximum - FCurrentLevel Then
        Begin
        Found:=TRUE;
        End
      Else
        Begin
        Comp:=FClaimQueue.Predecessor(Comp);
        End;
      End;
    End;
  Result:=Comp;
  End;   {LIFOSelect}

Function SCCSelect: TomasElement;
  Var
    SComp: TomasElement;
    NComp: TomasElement;
  Begin
  With _CurrentSemaphore Do
    Begin
    SComp:=FClaimQueue.FirstElement;
    NComp:=SComp;
    While NComp <> Nil Do
      Begin
      If ClaimSizeOf(NComp) < ClaimSizeOf(SComp) Then
        Begin
        SComp:=NComp;
        End;
      NComp:=FClaimQueue.Successor(NComp);
      End;
    If (SComp <> Nil) And (ClaimSizeOf(SComp) > FMaximum - FCurrentLevel) Then
      Begin
      SComp:=Nil;
      End;
    End;
  Result:=SComp;
  End;   {TomasSemaphore.SCCSelect}

Function GCCSelect: TomasElement;
  Var
    GComp: TomasElement;
    NComp: TomasElement;
  Begin
  With _CurrentSemaphore Do
    Begin
    GComp:=FClaimQueue.FirstElement;
    NComp:=GComp;
    While NComp <> Nil Do
      Begin
      If ClaimSizeOf(GComp) < ClaimSizeOf(NComp) Then
        Begin
        GComp:=NComp;
        End;
      NComp:=FClaimQueue.Successor(NComp);
      End;
    If (GComp <> Nil) And (ClaimSizeOf(GComp) > FMaximum - FCurrentLevel) Then
      Begin
      GComp:=Nil;
      End;
    End;
  Result:=GComp;
  End;   {GCCSelect}



end.
